'use strict';
Object.defineProperty(exports, "__esModule", { value: true });
const bignumber_js_1 = require("bignumber.js");
const BALANCE = (a) => a;
const INCOMING_CLAIM = (a) => a + ':claim';
const CHANNEL = (a) => a + ':channel';
const IS_BLOCKED = (a) => a + ':block';
const CLIENT_CHANNEL = (a) => a + ':client_channel';
const OUTGOING_BALANCE = (a) => a + ':outgoing_balance';
const LAST_CLAIMED = (a) => a + ':last_claimed';
class Account {
    constructor(opts) {
        this._store = opts.store;
        this._account = opts.account;
        this._api = opts.api;
        this._currencyScale = opts.currencyScale;
        this._funding = false;
        this._log = opts.log;
    }
    xrpToBase(amount) {
        return new bignumber_js_1.default(amount)
            .times(Math.pow(10, this._currencyScale))
            .toString();
    }
    getAccount() {
        return this._account;
    }
    getPaychan() {
        return this._paychan;
    }
    getClientPaychan() {
        return this._clientPaychan;
    }
    setClaimIntervalId(claimIntervalId) {
        this._claimIntervalId = claimIntervalId;
    }
    getClaimIntervalId() {
        return this._claimIntervalId;
    }
    getLastClaimedAmount() {
        return this._store.getString(LAST_CLAIMED(this._account)) || '0';
    }
    setLastClaimedAmount(amount) {
        this._store.set(LAST_CLAIMED(this._account), amount);
    }
    isFunding() {
        return this._funding;
    }
    setFunding(funding) {
        this._funding = funding;
    }
    async connect() {
        await Promise.all([
            this._store.loadString(BALANCE(this._account)),
            this._store.loadObject(INCOMING_CLAIM(this._account)),
            this._store.loadString(CHANNEL(this._account)),
            this._store.loadString(IS_BLOCKED(this._account)),
            this._store.loadString(CLIENT_CHANNEL(this._account)),
            this._store.loadString(OUTGOING_BALANCE(this._account)),
            this._store.loadString(LAST_CLAIMED(this._account))
        ]);
        const channelId = this.getChannel();
        if (channelId) {
            try {
                const paychan = await this._api.getPaymentChannel(channelId);
                this._paychan = paychan;
                this.setLastClaimedAmount(this.xrpToBase(paychan.balance));
            }
            catch (e) {
                this._log.error('failed to load channel entry. error=' + e.message);
                if (e.name === 'RippledError' && e.message === 'entryNotFound') {
                    this._log.error('removing channel because it has been deleted');
                    this.deleteChannel();
                }
            }
        }
        const clientChannelId = this.getClientChannel();
        if (clientChannelId) {
            this._clientPaychan = await this._api.getPaymentChannel(clientChannelId)
                .catch(() => ({}));
        }
    }
    async disconnect() {
        this._store.unload(BALANCE(this._account));
        this._store.unload(INCOMING_CLAIM(this._account));
        this._store.unload(CHANNEL(this._account));
        this._store.unload(IS_BLOCKED(this._account));
        this._store.unload(CLIENT_CHANNEL(this._account));
        this._store.unload(OUTGOING_BALANCE(this._account));
    }
    getBalance() {
        return new bignumber_js_1.default(this._store.getString(BALANCE(this._account)) || '0');
    }
    getIncomingClaim() {
        const paychanAmount = new bignumber_js_1.default(this.getLastClaimedAmount());
        const storedClaim = this._store.getObject(INCOMING_CLAIM(this._account)) ||
            { amount: '0' };
        if (paychanAmount.gt(storedClaim.amount)) {
            return { amount: paychanAmount.toString() };
        }
        else {
            return storedClaim;
        }
    }
    getChannel() {
        return this._store.getString(CHANNEL(this._account));
    }
    isBlocked() {
        return this._store.getString(IS_BLOCKED(this._account)) === 'true';
    }
    getClientChannel() {
        return this._store.getString(CLIENT_CHANNEL(this._account));
    }
    getOutgoingBalance() {
        return new bignumber_js_1.default(this._store.getString(OUTGOING_BALANCE(this._account)) || '0');
    }
    setBalance(balance) {
        return this._store.set(BALANCE(this._account), balance);
    }
    setIncomingClaim(incomingClaim) {
        return this._store.set(INCOMING_CLAIM(this._account), incomingClaim);
    }
    setChannel(channel, paychan) {
        this._paychan = paychan;
        this.setLastClaimedAmount(this.xrpToBase(paychan.balance));
        return this._store.set(CHANNEL(this._account), channel);
    }
    deleteChannel() {
        if (new bignumber_js_1.default(this.getLastClaimedAmount()).lt(this.getIncomingClaim().amount)) {
            this._log.error('Critical Error! Full balance was not able to be claimed before channel deletion.' +
                ' claim=' + JSON.stringify(this._store.getObject(INCOMING_CLAIM(this._account))) +
                ' lastClaimedAmount=' + this.getLastClaimedAmount() +
                ' channelId=' + this._store.getString(CHANNEL(this._account)));
        }
        const newBalance = new bignumber_js_1.default(this.getBalance())
            .minus(this.getLastClaimedAmount())
            .toString();
        this.setBalance(newBalance);
        delete this._paychan;
        this._store.delete(LAST_CLAIMED(this._account));
        this._store.delete(INCOMING_CLAIM(this._account));
        return this._store.delete(CHANNEL(this._account));
    }
    block(isBlocked = true) {
        return this._store.set(IS_BLOCKED(this._account), String(isBlocked));
    }
    async setClientChannel(clientChannel) {
        this._clientPaychan = await this._api.getPaymentChannel(clientChannel);
        return this._store.set(CLIENT_CHANNEL(this._account), clientChannel);
    }
    setOutgoingBalance(outgoingBalance) {
        return this._store.set(OUTGOING_BALANCE(this._account), outgoingBalance);
    }
}
exports.default = Account;
//# sourceMappingURL=account.js.map